//
// Skirmish Base Script.
//
// contains the rules for starting and ending a game.
// as well as warning messages.
//
// /////////////////////////////////////////////////////////////////

// Rules.js has been modified by WZM to match contingency.
// The tier limiter is now deprecated.
// This rules.js file is highly recommended to use on contingency.
// The readme in the root of the mod directory is what you should
// Read first to know what is changed.

receiveAllEvents(true);  // If doing this in eventGameInit, it seems to be too late in T2/T3, due to some eventResearched events triggering first.

include("multiplay/script/camTechEnabler.js");
include("multiplay/script/weather.js");

// We'll be using these arrays to ensure that the sound effects that we'll be using won't flood the player's ears with redundant info. ;)
var lastDetectTime = [0,0,0,0,0,0,0,0,0];
var lastHitTime = [0,0,0];
var lastDestroyedTime = [0,0,0,0,0];
var powerTimer = [0,0];
var playerDefeated = [0,0,0,0,0,0,0,0,0,0,0,0];

//This mod makes usage of alternate means of getting power, such as wind farms. Players also receive 1 free unit of power every second, so even if they don't have any Oil Derricks, they'll still receive a small trickle of power with which to build new stuff.
var additionalPower = [0,0,0,0,0,0,0,0,0,0,0,0];

var lastHitTime = 0;
var cheatmode = false;
var maxOilDrums = 0;
var mainReticule = true;
var allowDesign = false;

const CREATE_LIKE_EVENT = 0;
const DESTROY_LIKE_EVENT = 1;
const TRANSFER_LIKE_EVENT = 2;

function setMainReticule()
{
	setReticuleButton(0, _("Close"), "image_cancel_up.png", "image_cancel_down.png");
	if (countStruct("A0LightFactory") + countStruct("A0CyborgFactory") + countStruct("A0VTolFactory1") > 0)
	{
		setReticuleButton(1, _("Manufacture (F1)"), "image_manufacture_up.png", "image_manufacture_down.png");
	}
	else
	{
		setReticuleButton(1, _("Manufacture - build factory first"), "", "");
	}
	if (countStruct("A0ResearchFacility") > 0)
	{
		setReticuleButton(2, _("Research (F2)"), "image_research_up.png", "image_research_down.png");
	}
	else
	{
		setReticuleButton(2, _("Research - build research facility first"), "", "");
	}
	if (countDroid(DROID_CONSTRUCT, selectedPlayer) > 0)
	{
		setReticuleButton(3, _("Build (F3)"), "image_build_up.png", "image_build_down.png");
	}
	else
	{
		setReticuleButton(3, _("Build - manufacture constructor droids first"), "", "");
	}
	if (allowDesign == true)
	{
		setReticuleButton(4, _("Design (F4)"), "image_design_up.png", "image_design_down.png");
	}
	else
	{
		setReticuleButton(4, _("Design - construct HQ first"), "", "");
	}
	setReticuleButton(5, _("Intelligence Display (F5)"), "image_intelmap_up.png", "image_intelmap_down.png");
	if (countDroid(DROID_COMMAND, selectedPlayer) > 0 && countStruct("A0ComDroidControl") > 0)
	{
		setReticuleButton(6, _("Commanders (F6)"), "image_commanddroid_up.png", "image_commanddroid_down.png");
	}
	else
	{
		setReticuleButton(6, _("Commanders - manufacture commanders first"), "", "");
	}
	mainReticule = true; // main reticule window is open
}

function reticuleUpdate(obj, eventType)
{
	var update_reticule = false;
	var tryUpdate = (obj.player === selectedPlayer || eventType === TRANSFER_LIKE_EVENT);

	if (tryUpdate && obj.type === STRUCTURE)
	{
		if (obj.stattype === HQ)
		{
			var flag = (countStruct("A0CommandCentre", selectedPlayer) > 0);
			setMiniMap(flag); // show minimap or not
			setDesign(flag); // permit designs or not
			allowDesign = flag;
			update_reticule = true;
		}

		if (obj.stattype === RESEARCH_LAB || obj.stattype === CYBORG_FACTORY ||
			obj.stattype === VTOL_FACTORY || obj.stattype === FACTORY || obj.stattype === COMMAND_CONTROL)
		{
			update_reticule = true;
		}
	}
	else if (tryUpdate && obj.type === DROID)
	{
		if (obj.droidType === DROID_CONSTRUCT || obj.droidType === DROID_COMMAND)
		{
			update_reticule = true;
		}
	}

	if (mainReticule && update_reticule)
	{
		//Wait a tick for the counts to update
		const TICK_TIME = 100;
		queue("setMainReticule", TICK_TIME);
	}
}

function setupGame()
{
	if (tilesetType == "URBAN")
	{
		replaceTexture("page-8-player-buildings-bases.png", "page-8-player-buildings-bases-urban.png");
		replaceTexture("page-9-player-buildings-bases.png", "page-9-player-buildings-bases-urban.png");
		replaceTexture("page-7-barbarians-arizona.png", "page-7-barbarians-urban.png");
	}
	else if (tilesetType == "ROCKIES")
	{
		replaceTexture("page-8-player-buildings-bases.png", "page-8-player-buildings-bases-rockies.png");
		replaceTexture("page-9-player-buildings-bases.png", "page-9-player-buildings-bases-rockies.png");
		replaceTexture("page-7-barbarians-arizona.png", "page-7-barbarians-kevlar.png");
		// for some reason rockies will use arizona babas
	}
	if (tilesetType != "ARIZONA")
	{
		setSky("texpages/page-25-sky-urban.png", 0.5, 10000.0);
	}
	setMainReticule();
	showInterface();
	hackPlayIngameAudio();
}

function eventGameLoaded()
{
	setupGame();
}

function eventGameInit()
{
	setupGame();

	// always at least one oil drum, and one more for every 64x64 tiles of map area
	maxOilDrums = (mapWidth * mapHeight) >> 12; // replace float division with shift for sync-safety
	for (var i = 0; i < maxOilDrums; ++i)
	{
		queue("placeOilDrum", 10000 * i);
	}


	hackNetOff();
	makeComponentAvailable("B4body-sml-trike01", scavengerPlayer);
	makeComponentAvailable("B3body-sml-buggy01", scavengerPlayer);
	makeComponentAvailable("B2JeepBody", scavengerPlayer);
	makeComponentAvailable("BusBody", scavengerPlayer);
	makeComponentAvailable("FireBody", scavengerPlayer);
	makeComponentAvailable("B1BaBaPerson01", scavengerPlayer);
	makeComponentAvailable("BaBaProp", scavengerPlayer);
	makeComponentAvailable("BaBaLegs", scavengerPlayer);
	makeComponentAvailable("bTrikeMG", scavengerPlayer);
	makeComponentAvailable("BuggyMG", scavengerPlayer);
	makeComponentAvailable("BJeepMG", scavengerPlayer);
	makeComponentAvailable("BusCannon", scavengerPlayer);
	makeComponentAvailable("BabaFlame", scavengerPlayer);
	makeComponentAvailable("BaBaMG", scavengerPlayer);
	setPower(250, scavengerPlayer);
	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		if (powerType == 0)
		{
			setPower(1250, playnum);	// everything else is 1250 power
			setPowerModifier(100, playnum);
		}
		else if (powerType == 2)
		{
			setPower(1250, playnum);	// everything else is 1250 power
			setPowerModifier(145, playnum);
		}

		// insane difficulty is meant to be insane...
		if (playerData[playnum].difficulty == INSANE)
		{
			setPower(2500, playnum);	// insane is 2500 power
			setPowerModifier(180, playnum);
		}
		else if (playerData[playnum].difficulty == EASY)
		{
			setPower(1250, playnum);	// everything else is 1250 power
			setPowerModifier(90, playnum);
		}
		else if (playerData[playnum].difficulty == HARD)
		{
			setPower(1250, playnum);	// now 1250 regardless of dificulty
			setPowerModifier(155, playnum);
		}

		setDroidLimit(playnum, 400, DROID_ANY);	// droid limit for players
		setDroidLimit(playnum, 40, DROID_COMMAND);
		setDroidLimit(playnum, 40, DROID_CONSTRUCT);

		enableStructure("A0CommandCentre", playnum);		// make structures available to build
		enableStructure("A0LightFactory", playnum);
		enableStructure("A0ResourceExtractor", playnum);
		enableStructure("A0PowerGenerator", playnum);
		enableStructure("A0ResearchFacility", playnum);

		setStructureLimits("A0LightFactory", 10, playnum);	// set structure limits
		setStructureLimits("A0PowerGenerator", 64, playnum);
		setStructureLimits("A0ResearchFacility", 10, playnum);
		setStructureLimits("A0CommandCentre", 1, playnum);
		setStructureLimits("A0ComDroidControl", 128, playnum);
		setStructureLimits("A0CyborgFactory", 10, playnum);
		setStructureLimits("A0VTolFactory1", 10, playnum);
	}
	applyLimitSet();	// set limit options

	const numCleanTech = 4;	// do x for clean
	const numBaseTech = 18; // do x for base
	var techlist = new Array(
		"R-Vehicle-Prop-Wheels",
		"R-Sys-Spade1Mk1",
		"R-Vehicle-Body01",
		"R-Comp-SynapticLink",
		"R-Vehicle-Body04",
		"R-Wpn-MG1Mk1",
		"R-Defense-TankTrap01",
		"R-Defense-TankTrapGate",
		"R-Defense-TowerConst01",
		"R-Sys-Engineering01",
		"R-Vehicle-Prop-Halftracks",
		"R-Wpn-MG-Accuracy01",
		"R-Wpn-MG-Damage01",
		"R-Wpn-MG-ROF01",
		"R-Wpn-Cannon1Mk1",
		"R-Wpn-MG2Mk1",
		"R-Struc-RepairFacility",
		"R-Struc-PowerModuleMk1",
		"R-Wpn-MG3Mk1",
		"R-Wpn-Cannon-Accuracy01",
		"R-Wpn-Flamer01Mk1",
		"R-Wpn-Rocket05-MiniPod",
		"R-Wpn-Mortar01Lt",
		"R-Vehicle-Engine01",
		"R-Sys-Sensor-Turret01",
		"R-Sys-MobileRepairTurret01",
		"R-Comp-CommandTurret01",
		"R-Struc-CommandRelay",
		"R-Defense-EmplConst01",
		"R-Defense-HardcreteWall",
		"R-Defense-HardcreteGate",
		"R-Defense-TowerConst02",
		"R-Defense-BunkerConst01",
		"R-Defense-WallTowerConst01",
		"R-Struc-Research-Module",
		"R-Struc-Factory-Module",
		"R-Struc-Factory-Cyborg",
		"R-Cyborg-HvyConst",
		"R-Struc-Research-Upgrade01");

	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		enableResearch("R-Vehicle-Body04", playnum);
		enableResearch("R-Vehicle-Prop-Halftracks", playnum);
		enableResearch("R-Defense-TankTrap01", playnum);
		enableResearch("R-Wpn-MG1Mk1", playnum);
		enableResearch("R-Sys-Engineering01", playnum);
		enableResearch("R-Defense-TowerConst01", playnum);

		// enable cyborgs components that can't be enabled with research
		makeComponentAvailable("CyborgSpade", playnum);
		makeComponentAvailable("CyborgRepair", playnum);
		makeComponentAvailable("CyborgSpade2", playnum);
		makeComponentAvailable("CyborgRepair2", playnum);
		makeComponentAvailable("CyborgSpade3", playnum);
		makeComponentAvailable("CyborgRepair3", playnum);
		makeComponentAvailable("CyborgSpade4", playnum);
		makeComponentAvailable("CyborgRepair4", playnum);
		makeComponentAvailable("CyborgLegs", playnum);
		makeComponentAvailable("HvyCyborgLegs", playnum);
		makeComponentAvailable("Cyb-Wpn-Atmiss", playnum);
		makeComponentAvailable("CyborgCannon", playnum);
		makeComponentAvailable("CyborgChaingun", playnum);
		makeComponentAvailable("CyborgFlamer01", playnum);
		makeComponentAvailable("Cyb-Wpn-Grenade", playnum);
		makeComponentAvailable("Cyb-Hvywpn-A-T", playnum);
		makeComponentAvailable("Cyb-Hvywpn-Acannon", playnum);
		makeComponentAvailable("Cyb-Hvywpn-HPV", playnum);
		makeComponentAvailable("Cyb-Hvywpn-Mcannon", playnum);
		makeComponentAvailable("Cyb-Hvywpn-PulseLsr", playnum);
		makeComponentAvailable("Cyb-Hvywpn-RailGunner", playnum);
		makeComponentAvailable("Cyb-Hvywpn-TK", playnum);
		makeComponentAvailable("Cyb-Wpn-Laser", playnum);
		makeComponentAvailable("Cyb-Wpn-Rail1", playnum);
		makeComponentAvailable("CyborgRocket", playnum);
		makeComponentAvailable("CyborgRotMG", playnum);
		makeComponentAvailable("Cyb-Wpn-Thermite", playnum);
		makeComponentAvailable("CyborgFlamer01", playnum);
		makeComponentAvailable("HvyCybSpade", playnum);
		makeComponentAvailable("HvyCybRepair", playnum);
		makeComponentAvailable("HvyCybSpade2", playnum);
		makeComponentAvailable("HvyCybRepair2", playnum);
		makeComponentAvailable("HvyCybSpade3", playnum);
		makeComponentAvailable("HvyCybRepair3", playnum);
		makeComponentAvailable("HvyCybSpade4", playnum);
		makeComponentAvailable("HvyCybRepair4", playnum);
		makeComponentAvailable("Cyb-Wpn-MortarInc", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MortarRot", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Cannon-HPVAslt", playnum);
		makeComponentAvailable("Cyb-Wpn-Cannon-HPVLgt", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Cannon-Shotgun", playnum);
		makeComponentAvailable("Cyb-Wpn-Cannon-Sniper", playnum);
		makeComponentAvailable("HvyCyb-Wpn-EMP", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Flamer1", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Flamer2", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Flamer3", playnum);
		makeComponentAvailable("Cyb-Wpn-Laser-AP1", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Laser-AP1", playnum);
		makeComponentAvailable("Cyb-Wpn-Laser-AT1", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Laser-AT1", playnum);
		makeComponentAvailable("Cyb-Wpn-Laser-AP3", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Laser-AP3", playnum);
		makeComponentAvailable("Cyb-Wpn-Laser-AT3", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Laser-AT3", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG3", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG4", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG-AntiTank", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG-AntiTank2", playnum);
		makeComponentAvailable("Cyb-Wpn-MG3Inc", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG3Inc", playnum);
		makeComponentAvailable("Cyb-Wpn-MG4Inc", playnum);
		makeComponentAvailable("HvyCyb-Wpn-MG4Inc", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Railgun-Gattling", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Railgun-Shotgun", playnum);
		makeComponentAvailable("Cyb-Wpn-Railgun-Sniper", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Railgun-Aslt", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-Pod", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-Pod2", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-Pod3", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-LtA-T", playnum);
		makeComponentAvailable("Cyb-Wpn-Rocket-HvyA-T", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-MRL", playnum);
		makeComponentAvailable("Cyb-Wpn-Rocket-Sunburst", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-Sunburst", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Missile-MdArt", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Missile-LtSAM", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Missile-HvySAM", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-TopAttack", playnum);
		makeComponentAvailable("Cyb-Wpn-Rocket-Cherub", playnum);
		makeComponentAvailable("HvyCyb-Wpn-Rocket-Cherub", playnum);
		makeComponentAvailable("Cyb-Wpn-SpyTurret01", playnum);
		makeComponentAvailable("Cyb-Wpn-SpyTurret02", playnum);
		makeComponentAvailable("Cyb-Wpn-SpyTurret03", playnum);
		makeComponentAvailable("HvyCyb-Wpn-SpyTurretNEXUS", playnum);

		if (baseType == CAMP_CLEAN)
		{
			for (var count = 0; count < numCleanTech; count++)
			{
				completeResearch(techlist[count], playnum);
			}
			// Keep only some structures for insane AI
			var structs = enumStruct(playnum);
			for (var i = 0; i < structs.length; i++)
			{
				var s = structs[i];
				if (playerData[playnum].difficulty != INSANE
				    || (s.stattype != WALL && s.stattype != DEFENSE && s.stattype != GATE
				        && s.stattype != RESOURCE_EXTRACTOR))
				{
					removeObject(s, false);
				}
			}
		}
		else if (baseType == CAMP_BASE)
		{
			for (var count = 0; count < numBaseTech; count++)
			{
				completeResearch(techlist[count], playnum);
			}
			// Keep only some structures
			var structs = enumStruct(playnum);
			for (var i = 0; i < structs.length; i++)
			{
				var s = structs[i];
				if ((playerData[playnum].difficulty != INSANE && (s.stattype == WALL || s.stattype == DEFENSE))
				    || s.stattype == GATE || s.stattype == CYBORG_FACTORY || s.stattype == COMMAND_CONTROL)
				{
					removeObject(s, false);
				}
			}
		}
		else // CAMP_WALLS
		{
			for (var count = 0; count < techlist.length; count++)
			{
				completeResearch(techlist[count], playnum);
			}
		}
	}

	var techLevel = getMultiTechLevel();
	if (techLevel == 2)
	{
		grantTech(TECH_TWO);
	}
	else if (techLevel == 3)
	{
		grantTech(TECH_THREE);
	}

	// Disabled by default
	setMiniMap(false);
	setDesign(false);
	allowDesign = false;
	// This is the only template that should be enabled before design is allowed
	enableTemplate("ConstructionDroid");

	var structlist = enumStruct(selectedPlayer, HQ);
	for (var i = 0; i < structlist.length; i++)
	{
		// Simulate build events to enable minimap/unit design when an HQ exists
		eventStructureBuilt(structlist[i]);
	}

	hackNetOn();
	setTimer("checkEndConditions", 3000);
	if (tilesetType === "URBAN" || tilesetType === "ROCKIES")
	{
		setTimer("weatherCycle", 45000);
	}
}

// /////////////////////////////////////////////////////////////////
// END CONDITIONS
function checkEndConditions()
{
	var factories = countStruct("A0LightFactory") + countStruct("A0CyborgFactory");
	var droids = countDroid(DROID_ANY);

	// Losing Conditions
	if (droids == 0 && factories == 0)
	{
		var gameLost = true;
		if (playerDefeated[selectedPlayer] == 0) //Inform the player that he/she is out of the game.
		{
			playerDefeated[selectedPlayer] = 1;
			playSound("pcv470.ogg");
		}

		/* If teams enabled check if all team members have lost  */
		if (alliancesType == ALLIANCES_TEAMS || alliancesType == ALLIANCES_UNSHARED)
		{
			for (var playnum = 0; playnum < maxPlayers; playnum++)
			{
				if (playnum != selectedPlayer && allianceExistsBetween(selectedPlayer, playnum))
				{
					factories = countStruct("A0LightFactory", playnum) + countStruct("A0CyborgFactory", playnum);
					droids = countDroid(DROID_ANY, playnum);
					if (droids > 0 || factories > 0)
					{
						gameLost = false;	// someone from our team still alive
						break;
					}
				}
			}
		}

		if (gameLost)
		{
			gameOverMessage(false);
			removeTimer("checkEndConditions");
			return;
		}
	}

	// Winning Conditions
	var gamewon = true;

	// check if all enemies defeated
	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		if (playnum != selectedPlayer && !allianceExistsBetween(selectedPlayer, playnum))	// checking enemy player
		{
			factories = countStruct("A0LightFactory", playnum) + countStruct("A0CyborgFactory", playnum); // nope
			droids = countDroid(DROID_ANY, playnum);
			if (droids == 0 && factories == 0)
			{
				if (playerDefeated[playnum] == 0) //Inform the player that an enemy player has been defeated.
				{
					playerDefeated[playnum] = 1;
					playSound("pcv646.ogg");
				}
			}
			else if (droids > 0 || factories > 0)
			{
				gamewon = false;	//one of the enemies still alive
				break;
			}
		}
	}

	if (gamewon)
	{
		gameOverMessage(true);
		removeTimer("checkEndConditions");
	}
}

// /////////////////////////////////////////////////////////////////
// Give all players a little power every second. More power is given if the player built more wind farms. INSANE AIs receive more power from Wind Turbines than other players, as well as additional power from each Oil Derrick they own.

function givePower()
{
   for (var playnum = 0; playnum < maxPlayers; playnum++)
   {
      var windFarms = enumStruct(playnum, "WindFarm");
      var activeWindFarms = 0;
      var activeDerricks = 0;
      var powerMultiplier = powerType/400;
      for (var i = 0; i < windFarms.length; i++)
      {
         if (windFarms[i].status === BUILT)
         {
            activeWindFarms += 1;
         }
      }
      if (playerData[playnum].difficulty === INSANE)
      {
         var Derricks = enumStruct(playnum, "A0ResourceExtractor");
         for (var i = 0; i < Derricks.length; i++)
         {
            if (Derricks[i].status === BUILT)
            {
               activeDerricks += 1;
            }
         }
         additionalPower[playnum] += Math.floor(((activeWindFarms * 2) + activeDerricks + 20) * 0.1);
      }
      else
      {
         additionalPower[playnum] += Math.floor((activeWindFarms + 10) * 0.1);
      }
      if (additionalPower[playnum] >= 1)
      {
         setPower(playerPower(playnum) + additionalPower[playnum], playnum);
         additionalPower[playnum] = 1;
      }
   }
}

// /////////////////////////////////////////////////////////////////
// WARNING MESSAGES
// Base Under Attack
function eventObjectSeen(viewer, detectedObj)
{
	if (detectedObj.player != selectedPlayer && allianceExistsBetween(selectedPlayer,detectedObj.player) != true && (viewer.player == selectedPlayer || allianceExistsBetween(selectedPlayer,viewer.player))) //Does the viewer belong to either the player or one of his/her allies, and is the detected object an enemy object?
	{
		if (detectedObj.type == DROID) //Enemy unit detected
		{
			if (detectedObj.player == scavengerPlayer) //Scavenger unit detected
			{
				if (gameTime > lastDetectTime[0] + 3000)
				{
					playSound("pcv373.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[0] = gameTime;
			}
			else if (detectedObj.droidType == DROID_TRANSPORTER || detectedObj.droidType == DROID_SUPERTRANSPORTER) //Enemy transport detected
			{
				if (gameTime > lastDetectTime[1] + 3000)
				{
					playSound("pcv388.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[1] = gameTime;
			}
			else if (detectedObj.isVTOL) //Enemy VTOL detected
			{
				if (gameTime > lastDetectTime[2] + 3000)
				{
					playSound("pcv388.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[2] = gameTime;
			}
			else if (detectedObj.hasIndirect && detectedObj.range > 2000) //Enemy artillery detected. NOTE: We need to exclude the more short-ranged indirect-fire units, like Mini-Rocket Arrays.
			{
				if (gameTime > lastDetectTime[3] + 3000)
				{
					playSound("pcv387.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[3] = gameTime;
			}
			else
			{
				if (gameTime > lastDetectTime[4] + 3000) //Play generic enemy unit detected sound
				{
					playSound("pcv378.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[4] = gameTime;
			}
		}
		else if (detectedObj.type == STRUCTURE && detectedObj.stattype != WALL && detectedObj.stattype != GATE) //Enemy structure detected (excluding walls or gates)
		{
			if (detectedObj.player == scavengerPlayer && !(detectedObj.hasIndirect && detectedObj.range > 2000)) //Scavenger structure detected
			{
				if (detectedObj.stattype == DEFENSE) //Scavenger defense detected
				{
					if (gameTime > lastDetectTime[5] + 3000)
					{
						playSound("pcv375.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
					}
					lastDetectTime[5] = gameTime;
				}
				else //Scavenger base detected
				{
					if (gameTime > lastDetectTime[6] + 3000)
					{
						playSound("pcv374.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
					}
					lastDetectTime[6] = gameTime;
				}
			}
			else if (detectedObj.hasIndirect && detectedObj.range > 2000) //Enemy artillery detected. NOTE: We need to exclude the more short-ranged indirect-fire units, like Mini-Rocket Arrays.
			{
				if (gameTime > lastDetectTime[7] + 3000)
				{
					playSound("pcv387.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[7] = gameTime;
			}
			else //Enemy player's structure detected
			{
				if (gameTime > lastDetectTime[8] + 3000)
				{
					playSound("pcv379.ogg", detectedObj.x, detectedObj.y, detectedObj.z);
				}
				lastDetectTime[8] = gameTime;
			}
		}
	}
}

// Under Attack
function eventAttacked(victimObj, attackerObj)
{
	if (victimObj.type == STRUCTURE) //Structure under attack
	{
		if (victimObj.stattype == RESOURCE_EXTRACTOR && gameTime > lastHitTime[0] + 5000) //Derrick under attack
		{
			lastHitTime[0] = gameTime;
			playSound("pcv345.ogg", victimObj.x, victimObj.y, victimObj.z);	// show position if still alive
		}
		else if (gameTime > lastHitTime[1] + 5000)//Generic structure under attack
		{
			lastHitTime[1] = gameTime;
			playSound("pcv337.ogg", victimObj.x, victimObj.y, victimObj.z);
		}
	}
	else if (victimObj.type == DROID && gameTime > lastHitTime[2] + 5000) //Unit under attack
	{
		lastHitTime[2] = gameTime;
		playSound("pcv399.ogg", victimObj.x, victimObj.y, victimObj.z);
	}
}

function eventStructureBuilt(struct)
{
	if (struct.player === selectedPlayer)
	{
		reticuleUpdate(struct, CREATE_LIKE_EVENT);
	}
}

function eventDestroyed(victim)
{
	if (victim.player == selectedPlayer)
	{
		if (victim.type == DROID && gameTime > lastDestroyedTime[0] + 3000) //Unit destroyed
		{
			playSound("pcv400.ogg", victim.x, victim.y, victim.z);
			lastDestroyedTime[0] = gameTime;
		}
		else if (victim.type == STRUCTURE)
		{
			if (victim.stattype == RESOURCE_EXTRACTOR && gameTime > lastDestroyedTime[1] + 3000) //Oil derrick destroyed
			{
				playSound("pcv346.ogg", victim.x, victim.y, victim.z);
				lastDestroyedTime[1] = gameTime;
			}
			else if (victim.stattype == POWER_GEN && gameTime > lastDestroyedTime[2] + 3000) //Power generator destroyed
			{
				playSound("pcv342.ogg", victim.x, victim.y, victim.z);
				lastDestroyedTime[2] = gameTime;
			}
			else if ((victim.stattype == LASSAT || victim.name == "Laser Satellite Command Post") && gameTime > lastDestroyedTime[3] + 3000) //LasSat Command Post destroyed
			{
				playSound("pcv651a.ogg", victim.x, victim.y, victim.z);
				lastDestroyedTime[3] = gameTime;
			}
			else
			{
				if (gameTime > lastDestroyedTime[4] + 3000) //Play generic structure destroyed sound
				{
					playSound("pcv620.ogg", victim.x, victim.y, victim.z);
					lastDestroyedTime[4] = gameTime;
				}
			}
		}
	}
}

// /////////////////////////////////////////////////////////////////
// Display miscellaneous situations to player that cannot be called by other known events.

function sitRep()
{
	if (playerPower(selectedPlayer) < 1 && gameTime > powerTimer[0]+15000)
	{
		playSound("pcv343.ogg"); //Warn the player that his/her power is low.
		powerTimer[0] = gameTime;
	}
	if (enumStruct(selectedPlayer,RESOURCE_EXTRACTOR).length > enumStruct(selectedPlayer,POWER_GEN).length*4 && gameTime > powerTimer[1]+15000)
	{
		playSound("pcv349.ogg"); //Warn the player that he/she needs more power generators.
		powerTimer[1] = gameTime;
	}
}

function eventDroidBuilt(droid, structure)
{
	if (droid.player === selectedPlayer)
	{
		reticuleUpdate(droid, CREATE_LIKE_EVENT);
	}
}

function eventObjectTransfer(obj, from)
{
	reticuleUpdate(obj, TRANSFER_LIKE_EVENT);
}

function eventResearched(research, structure, player)
{
	//if (research.name == "") debug("RESEARCH : " + research.fullname + "(" + research.name + ") for " + player);
	// iterate over all results
	for (var i = 0; i < research.results.length; i++)
	{
		var v = research.results[i];
		//if (research.name == "") debug("    RESULT : class=" + v['class'] + " parameter=" + v['parameter'] + " value=" + v['value'] + " filter=" + v['filterParameter'] + " filterval=" + v['filterValue']);
		for (var cname in Upgrades[player][v['class']]) // iterate over all components of this type
		{
			var parameter = v['parameter'];
			var ctype = v['class'];
			var filterparam = v['filterParameter'];
			if ('filterParameter' in v && Stats[ctype][cname][filterparam] != v['filterValue']) // more specific filter
			{
				//if (research.name == "") debug("    skipped param=" + parameter + " cname=" + cname);
				continue;
			}
			if (Stats[ctype][cname][parameter] instanceof Array)
			{
				var dst = Upgrades[player][ctype][cname][parameter].slice()
				for (var x = 0; x < dst.length; x++)
				{
					dst[x] += Math.ceil(Stats[ctype][cname][parameter][x] * v['value'] / 100);
				}
				Upgrades[player][ctype][cname][parameter] = dst
				//debug("    upgraded to " + dst);
			}
			else if (Stats[ctype][cname][parameter] > 0) // only applies if stat has above zero value already
			{
				Upgrades[player][ctype][cname][parameter] += Math.ceil(Stats[ctype][cname][parameter] * v['value'] / 100);
				//if (research.name == "") debug("      upgraded " + cname + " to " + Upgrades[player][ctype][cname][parameter] + " by " + Math.ceil(Stats[ctype][cname][parameter] * v['value'] / 100));
			}
			//else if (research.name == "") debug("    passed " + Stats[ctype][cname][parameter] + " param=" + parameter + " cname=" + cname);
		}
	}
}

function eventCheatMode(entered)
{
	cheatmode = entered; // remember this setting
}

function eventChat(from, to, message)
{
	if (message == "bettertogether" && cheatmode)
	{
		for (var i in Upgrades[from].Brain)
		{
			if (Upgrades[from].Brain[i].BaseCommandLimit > 0) // is commander
			{
				Upgrades[from].Brain[i].BaseCommandLimit += 4;
				Upgrades[from].Brain[i].CommandLimitByLevel += 2;
				// you must set the thresholds this way, as an array, because of the clunky
				// way that this is implemented behind the scenes
				Upgrades[from].Brain[i].RankThresholds = [ 0, 2, 4, 8, 16, 24, 32, 48, 64 ];
			}
		}
		console("Made player " + from + "'s commanders SUPERIOR!");
	}
	if (message == "makesuperior" && cheatmode)
	{
		for (var i in Upgrades[from].Body)
		{
			if (Upgrades[from].Body[i].bodyClass === 'Droids' || Upgrades[from].Body[i].bodyClass === 'Cyborgs')
			{
				Upgrades[from].Body[i].HitPoints += 500;
				Upgrades[from].Body[i].HitPointPct += 100;
				Upgrades[from].Body[i].Armour += 500;
				Upgrades[from].Body[i].Thermal += 500;
				Upgrades[from].Body[i].Power += 500;
			}
		}
		console("Made player " + from + "'s units SUPERIOR!");
	}
}

function placeOilDrum()
{
	var drums = enumFeature(-1, "OilDrum").length;
	if (drums >= maxOilDrums)
	{
		return;
	}

	var x = syncRandom(mapWidth - 20) + 10;
	var y = syncRandom(mapHeight - 20) + 10;

	// see if the random position is valid
	var occupied = (enumRange(x, y, 2, ALL_PLAYERS, false).length > 0);
	var unreachable = true;
	for (var i = 0; i < maxPlayers; ++i)
	{
		if (propulsionCanReach("wheeled01", x, y, startPositions[i].x, startPositions[i].y))
		{
			unreachable = false;
			break;
		}
	}
	var terrain = terrainType(x, y);
	if (terrain == TER_WATER || terrain == TER_CLIFFFACE)
	{
		unreachable = true;
	}
	if (occupied || unreachable)
	{
		// try again in a different position after 1 second
		queue("placeOilDrum", 1000);
		return;
	}

	addFeature("OilDrum", x, y);
}

function eventPickup(feature, droid)
{
	if (feature.stattype == OIL_DRUM)
	{
		var delay;
		// generate Geom(1/6) distribution for oil drum respawn delay
		for (delay = 0; ; ++delay)
		{
			if (syncRandom(6) == 0)
			{
				break;
			}
		}
		// amounts to 10 minutes average respawn time
		queue("placeOilDrum", delay * 120000);
	}
}
